""" file which stores the addon keymap and the  internal modal
    keymap and has functions to check for conflicts with it """
from bpy.types import KeyMapItems

user_view_keymaps = {}
user_others_keymaps = {}

# the View operator which we scan for user prefs
view_ops_names = [
    "Pan View",
    "Rotate View",
    "Zoom View",
    "Center View to Mouse",
    "Center View to Cursor",
    "View Axis",
    "View Orbit",
    "View Roll",
    "Pan View Direction"
]

other_ops_names = [
    #"Set Object Mode",
    "Toggle Shading Type",
    "Toggle X-Ray"
]

addon_keymap = None
addon_keys = []

modal_keymap = None
modal_keys = []

scanned = False

addon_shortcut_kmi = None


def create_keymaps(context):
    global modal_keymap
    global addon_keymap
    global addon_shortcut_kmi

    # addon keymap
    wm = context.window_manager
    addon_keyconfig = wm.keyconfigs.addon
    kc = addon_keyconfig

    addon_keymap = kc.keymaps.new(name="3D View", space_type="VIEW_3D")

    a_kmi = addon_keymap.keymap_items.new(
        # "object.face2face", 'F', 'RELEASE', shift=True)

        "object.face2face", 'F7', 'RELEASE')
    addon_keys.append(a_kmi)
    addon_shortcut_kmi = a_kmi
    # print('--------------------------keyMAP CREATED !!!!!!!!!!!!!!!!1--------------------------:')
    # modal keymap
    """ modal_keymap = [
        # KeyMapItems.new(idname, type, value, any=False, shift=False, ctrl=False, alt=False, oskey=False, key_modifier='NONE', head=False)
        KeyMapItems.new("face2Face.scale", 'S', 'RELEASE', shift=True),
        KeyMapItems.new("face2Face.grab", 'G', 'RELEASE', shift=True),
        KeyMapItems.new("face2Face.rotate", 'R', 'RELEASE', shift=True)
    ] """


view_on_middlemouse = False


def remove_keymaps(context):
    """ remove all keys when unregistering the addon """
    # print('--------------------------keyMAP REMOVE !!!!!!!!!!!!!!!!1--------------------------:')
    for kmi in addon_keys:
        addon_keymap.keymap_items.remove(kmi)
    addon_keys.clear()
    """ for kmi in modal_keys:
        modal_keymap.keymap_items.remove(kmi)
    modal_keys.clear() """


def scan_user_mapping(context):
    """ scan for user prefs hotkeys for the View operators
        - at the moment we scan for the actual settings and pass them
          through at the end of the modal event. When there are conflicts
          the user has no other choice than to change its settings
        - for better support we may could do:
        1:if the settings conflicts with modal map deactivate user settings temporary
          and use defaults
        2:implement routing to view ops from default settings for those who conflict
        3:allow to config modal map
    """

    global addon_shortcut_kmi
    # print('--------------------------keyMAP SCANNNNNN !!!!!!!!!!!!!!!!1--------------------------:')
    user_view_keymaps.clear()
    user_others_keymaps.clear()
    for kc_name, kc in context.window_manager.keyconfigs.items():
        # print('--------------------------KC NAME:',
        #      kc_name, " maps", kc.keymaps)
        # for kc_map in kc.keymaps:
        # print('----------KC MAP NAME:', kc_map.name)
        if "3D View" in kc.keymaps:
            # print('--------------------------KC NAME:', kc_name)
            for kmi in kc.keymaps["3D View"].keymap_items:
                # print('3D VIEW NAME:', kmi.name)
                # print(i.name, i.ctrl, i.shift, i.alt, i.value, i.type)
                # print(' NAME:', kmi.name)
                if(kmi.name == "Align by faces modal"):

                    """ print('align by facesNAME:', kmi.name)
                    print('align by faces:', kmi.type)
                    print('align by faces:', kmi.value)
                    print('align by faces:', kmi.ctrl)
                    print('align by faces:', kmi.alt) """
                    addon_shortcut_kmi = kmi
                check_and_add_view_mapping(kmi)
                check_and_add_others_mapping(kmi)

        if "Object Non-modal" in kc.keymaps:
            #print('--------------------------Object non-modalKC NAME:', kc_name)
            for kmi in kc.keymaps["Object Non-modal"].keymap_items:
                check_and_add_others_mapping(kmi)
                # print('USER VIEW KEYMAP', user_view_keymaps)
                # print('USER VIEW KEYMAP len ops', len(user_view_keymaps))
    return False


def check_and_add_view_mapping(kmi):
    global view_on_middlemouse
    global user_view_keymaps
    for name in view_ops_names:

        if(kmi.name == name):
            if(kmi.name == "Rotate View"):

                """ print('3D VIEW NAME:', kmi.name)
                print('USER VIEW KEY:', kmi.type)
                print('USER VIEW KEY:', kmi.value)
                print('USER VIEW KEY:', kmi.ctrl)
                print('USER VIEW KEY:', kmi.alt) """
                if kmi.type == "MIDDLEMOUSE":
                    view_on_middlemouse = True
                    # print('3D VIEW NAME:!!!!!!!!!!!!!!!!!')

            if name in user_view_keymaps:
                user_view_keymaps[name].append(kmi)
            else:
                user_view_keymaps[name] = [kmi]
            # print('USER VIEW KEYMAP len', len(
            #    user_view_keymaps[name]))


def check_and_add_others_mapping(kmi):
    global user_others_keymaps
    for name in other_ops_names:
        # print('-------Object non-modal opsname:', name)
        # print('-------Object non-modal kminame:', kmi.name)
        if(kmi.name == name):

            if name in user_others_keymaps:
                user_others_keymaps[name].append(kmi)
            else:
                user_others_keymaps[name] = [kmi]


def is_user_relevant_key(event):
    return is_user_view_key(event) or is_user_others_key(event)


def is_user_view_key(event):
    """ check if event exists in user defined view keys """
    for op_map_key in user_view_keymaps:
        op_map = user_view_keymaps[op_map_key]
        for kmi in op_map:

            if key_macthes_event(kmi, event):
                return True

    return False


def is_user_others_key(event):
    """ check if event exists in user defined view keys """
    for op_map_key in user_others_keymaps:
        op_map = user_others_keymaps[op_map_key]
        for kmi in op_map:
            if key_macthes_event(kmi, event):
                return True

    return False


def key_macthes_event_value2(kmi, event):
    return (event.value == kmi.value
            and (kmi.value == "ANY"
                 or (kmi.value == "CLICK" and event.value == "RELEASE")
                 or (kmi.value == "CLICK" and event.value == "PRESS")))


def key_macthes_event_value(kmi, event):
    return (event.value == kmi.value
            or kmi.value == "ANY"
            or (kmi.value == "CLICK" and event.value == "RELEASE")
            or (kmi.value == "CLICK" and event.value == "PRESS"))


def key_macthes_event(kmi, event):
    return (key_macthes_event_value(kmi, event) and
            event.type == kmi.type and
            event.shift == kmi.shift and
            event.ctrl == kmi.ctrl and
            event.alt == kmi.alt)


def key_macthes_event_ole(kmi, event, op_map_key):
    # if(kmi.type == "MIDDLEMOUSE" and event.type == "MIDDLEMOUSE"):
    if(op_map_key == "Context Toggle" and event.value == "PRESS"):
        """ print("op_map.name", op_map_key)
        print("kmi.name", kmi.name)
        print("event.type == kmi.type", event.type, " == ", kmi.type)
        print("event.value == kmi.value", event.value, " == ", kmi.value, " ?", (event.value ==
                                                                                 kmi.value or kmi.value == "ANY" or (kmi.value == "CLICK" and event.value == "RELEASE")))
        print("event.shift == kmi.shift", event.shift, " == ", kmi.shift)
        print("event.ctrl == kmi.ctrl", event.ctrl, " == ", kmi.ctrl)
        print("event.alt == kmi.alt", event.alt, " == ", kmi.alt) """
    return (key_macthes_event_value(kmi, event) and
            event.type == kmi.type and
            event.shift == kmi.shift and
            event.ctrl == kmi.ctrl and
            event.alt == kmi.alt)


def conflicts_with_map(kmi):
    """ check if two keymap items conflict """
    for modal_kmi in modal_keymap:
        if conflicts(modal_kmi, kmi):
            return True

    return False


def conflicts(kmi1, kmi2):
    """ check if two keymap items conflict """
    if \
            kmi1.type == kmi2.type and \
            kmi1.ctrl == kmi2.ctrl and \
            kmi1.shift == kmi2.shift and \
            kmi1.oskey == kmi2.oskey and \
            kmi1.alt == kmi2.alt:

        return True

    return False


def determe_selection_mode(context, modal):
    """ determs if left or right click select or whatever
     (hints only work for left or right click) """
    keyconfigs = context.window_manager.keyconfigs

    if not 'blender' in keyconfigs and not 'Blender' in keyconfigs:
        print(
            "could not determe right or left click selection mode(left click choosen)!")
        # report will be overwritten by idle state TODO: think about solution
        # modal.report(
        #    {'WARNING'}, "could not determe right or left click selection mode(left click choosen)")
        return False

    blender_kc = keyconfigs['blender'] if 'blender' in keyconfigs else keyconfigs['Blender']

    for i in blender_kc.keymaps["3D View"].keymap_items:
        # for i in context.window_manager.keyconfigs['blender'].keymaps["3D View"].keymap_items:
        # for ii in i.keymap_items:
        # print(i.name, i.ctrl, i.shift, i.alt, i.value, i.type)
        if i.name == "Select":
            if i.type == 'RIGHTMOUSE':
                return True
    return False
