""" This file contains the operator to activate the face2Face addon"""
import bpy

# Import From Files
if "bpy" in locals():
    import importlib
    from . import helpers
    importlib.reload(helpers)
    from . import projection_helpers as pro_helpers
    importlib.reload(pro_helpers)
else:
    from . import projection_helpers as pro_helpers


class OBJECT_OT_Face2FaceProjectionCopy(bpy.types.Operator):
    bl_label = "copy face2Face projection - including all involved objects"
    bl_description = "Reactivate a face2Face projection by its name"
    bl_idname = "object.copy_face2face_projection"

    copy_projection_name: bpy.props.StringProperty(name='copy_projection_name',
                                                   default="NO_NAME")
    old_target_name = None

    old_target_obj = None
    new_objects = []

    def execute(self, context):
        print("copyig face2Face projection:" +
              self.copy_projection_name+"... -->")
        bpy.ops.object.mode_set(mode='OBJECT', toggle=False)

        if not self.select_objects_by_projection_name(context, self.copy_projection_name):
            return {'CANCELLED'}

        bpy.ops.object.duplicate_move()
        # bpy.ops.object.delete({"selected_objects": objects})
        context.view_layer.update()
        objects = context.selected_objects
        target_obj = context.active_object
        self.rename_projection_name(
            objects, self.old_target_obj.name, target_obj, context, self.copy_projection_name)
        context.view_layer.update()
        bpy.ops.ed.undo_push(message="face2Face copy")
        print("copyig face2Face projection finished:" +
              self.copy_projection_name+"... -->")
        return {'FINISHED'}

    def rename_projection_name(self, objects, old_target_obj_name, target_obj, context, projection_name, with_subs=True):
        print("old tn:", old_target_obj_name)
        print("old projn:", projection_name)
        # target_name =
        new_projection_name = projection_name.replace(
            old_target_obj_name, target_obj.name)

        print("new projn:", new_projection_name)

        copy_collection = bpy.data.collections.new(new_projection_name)
        context.scene.collection.children.link(copy_collection)

        for obj in objects:
            # new_objects.add(obj)
            if obj == target_obj:
                obj.projections.clear()
                new_projection = obj.projections.add()
                new_projection.projection_name = new_projection_name
            else:
                if obj.belongs_to == projection_name:
                    obj.belongs_to = new_projection_name
                    copy_collection.objects.link(obj)
                    if with_subs:
                        for projection in obj.projections:
                            self.set_target_object_by_projection_name(
                                context, projection.projection_name, obj)
                            self.rename_projection_name(
                                objects, self.old_target_obj.name, obj, context, projection.projection_name)

    def set_target_object_by_projection_name(self, context, projection_name, except_obj=None):
        """ because of copy there could tempoary exists two target object for same,
          before beeing renamed exclude the new one to get the old """
        for obj in context.scene.objects:
            if obj == except_obj:
                print("set_target_object_by_projection_name except_obj!!!!:", obj.name)
                continue
            if pro_helpers.is_target_object(obj, projection_name):
                print("set_target_object_by_projection_name set to", obj.name)
                self.old_target_obj = obj

    def select_objects_by_projection_name(self, context, projection_name, with_subs=True):
        """ determe and select involved objects of project and deselect all others  """
        objects = []
        self.set_target_object_by_projection_name(
            context, projection_name)

        objects = pro_helpers.get_objects_by_projection_name(
            context, projection_name, with_subs)

        if self.old_target_obj is None or len(objects) == 0:
            return False

        # select and deselect
        for obj in context.scene.objects:
            if obj in objects:
                obj.select_set(True)
            else:
                obj.select_set(False)

        self.old_target_obj.select_set(True)
        context.view_layer.objects.active = self.old_target_obj
        context.view_layer.update()

        return True


def register():
    bpy.utils.register_class(OBJECT_OT_Face2FaceProjectionCopy)


def unregister():
    bpy.utils.unregister_class(OBJECT_OT_Face2FaceProjectionCopy)
