""" just a few homeless helper methods """
import random
import math
import mathutils
import decimal
import bpy


def addon_prefs(context):
    if context is not None:
        return context.preferences.addons[__package__.partition('.')[0]].preferences
    return bpy.context.preferences.addons[__package__.partition('.')[0]].preferences


def flatten(mat):
    """ converts a matrix into a flat list(to strore in a FloatProperty) """
    dim = len(mat)
    return [mat[j][i] for i in range(dim)
            for j in range(dim)]


# math
def get_random_variation():
    return 1.5 - random.randrange(1, 100)/100


def get_random_rot_variation():
    return 2 * math.pi * random.randrange(1, 100)/100 - math.pi


def clamp(val, minimum, maximum):
    """ limit a value in min and max borders
    TODO: move somewhere else because not realy connect to mesh"""
    if val < minimum:
        return minimum
    if val > maximum:
        return maximum
    return val


def truncate(num, places):
    decimal.getcontext().rounding = decimal.ROUND_DOWN
    dec = decimal.Decimal(num)
    return float(round(dec, places))


def truncate2(number, digits) -> float:
    stepper = 10.0 ** digits
    return math.trunc(stepper * number) / stepper


changes = []


def update_children(context, obj, parent, linked=False,  copy_projections=True):
    """ fix childs matrix_parent_inverse which got screwed up by inner mesh rotation during align faces"""
    for child in obj.children:

        matrix_custom_transform_inv = parent.matrix_custom_transform.to_4x4()
        matrix_custom_transform_inv.invert()

        # insert inner mesh rotation into the childs matrix_parent_inverse
        child.matrix_parent_inverse = matrix_custom_transform_inv @ parent.inner_mesh_rot.to_4x4(
        ) @ parent.matrix_custom_transform.to_4x4() @ child.matrix_parent_inverse


def copy_obj(context, obj, parent, linked=False, with_children=True, copy_projections=True):
    """ copy object with new data or linked old data and with children or without """
    global changes
    changes = []
    copy = copy_obj_recursive(context, obj, parent,
                              linked, with_children, copy_projections)

    if copy_projections:
        change_obj_belongings(copy)

    return copy


def copy_obj_recursive(context, obj, parent, linked=False, with_children=True, copy_projections=True):
    """ copy object with new data or linked old data and with children or without """
    copy = obj.copy()
    if not linked:
        copy.data = obj.data.copy()
    if copy_projections:
        copy_projections_for(context, copy, obj.name)
    if not parent is None:
        copy.parent = parent
        copy.matrix_parent_inverse = obj.matrix_parent_inverse.copy()
    if with_children:
        for child in obj.children:
            copy_obj_recursive(context, child, copy, linked)
    return copy


def get_in_changes(name):
    global changes
    for change_tuple in changes:
        if change_tuple[0] == name:
            return change_tuple[1]
    return None


def change_obj_belongings(obj):
    global changes
    for child in obj.children:
        changed_name = get_in_changes(child.belongs_to)
        if not changed_name is None:
            child.belongs_to = changed_name
            change_obj_belongings(child)


def copy_projections_for(context, obj, oiginal_obj_name):
    global changes

    for projection in obj.projections:
        # print("old projn:", projection.projection_name)
        new_projection_name = projection.projection_name.replace(
            oiginal_obj_name, obj.name)

        # print("new projn:", new_projection_name)
        changes.append((projection.projection_name, new_projection_name))
        projection.projection_name = new_projection_name
        copy_collection = bpy.data.collections.new(new_projection_name)
        context.scene.collection.children.link(copy_collection)


def get_all_children(obj):
    """ select object and their children and subchildren all the way the hierachy down"""
    objects = []
    for child in obj.children:
        objects.append(child)
        objects += get_all_children(child)
    return objects


def select_obj(obj, with_children=True):
    """ select object and their children or without """
    obj.select_set(True)
    if with_children:
        for child in obj.children:
            select_obj(child)


def link_obj(obj, collection, with_children=True):
    """ link object to projection collection with children or without """
    collection.objects.link(obj)
    if with_children:
        for child in obj.children:
            link_obj(child, collection)
