""" represents the Modal state where object beeing moved """
import mathutils
import bpy

if "bpy" in locals():
    import importlib
    from .. import mesh
    importlib.reload(mesh)
    from .. import view
    importlib.reload(view)
    from . import base
    importlib.reload(base)
    from . base import ModalState
    from .. import helpers
    importlib.reload(helpers)
else:
    # pass
    from .. import mesh
    from .. import view
    from . import base
    from . base import ModalState
    from .. import helpers


class GrabState(ModalState):
    """ represents the Modal state where object beeing moved """
    name = "MOVE"
    # mouse_start = (0, 0)

    # if set individual target_face scale and form are take into account
    relative_pos = True

    # if set movement in limited to the borders of the target face
    # could only be set to false in absolute positioning mode
    limit_to_face = True
    inside = True  # indicates if face borders were respected
    axis_locked = 0

    @staticmethod
    def adapt_offset(modal):
        """ adopt offset from active to selected """
        active_obj = modal.projected_objects[modal.object_index]
        target_face = modal.target_object.data.polygons[active_obj.target_face_index]
        tverts = mesh.get_verts_from_polygon(modal.target_object, target_face)
        target_face_mid = mesh.get_midpoint_world(modal.target_object, tverts)
        source_face_mid  = mesh.get_midpoint_by_active(active_obj)
        active_offset = source_face_mid - target_face_mid

        matrix_custom_transform_inv = active_obj.matrix_custom_transform.copy()
        matrix_custom_transform_inv.invert()
        
        active_offset = matrix_custom_transform_inv @ active_offset
        
        for obj in modal.selected_objects:
            if obj is not active_obj:

                oactive_offset = obj.matrix_custom_transform @ active_offset.copy()

                otarget_face = modal.target_object.data.polygons[obj.target_face_index]
                otverts = mesh.get_verts_from_polygon(modal.target_object, otarget_face)
                otarget_face_mid = mesh.get_midpoint_world(modal.target_object, otverts)
                osource_face_mid = mesh.get_midpoint_by_active(obj)
                
                obj.location = otarget_face_mid + oactive_offset + (obj.location-osource_face_mid)
                
    
    def set_axis(self, context, event, modal):
        if event.type in {'X'}:
            if event.value == 'RELEASE':
                if self.axis_locked in (0, 2, 3):
                    self.axis_locked = 1
                    if context.window_manager.face2face.auto_view:
                        self.align_view2obj(context, modal)
                    modal.report({'INFO'}, "grabbing mode locked to x axis")
                else:
                    self.axis_locked = 0
                    modal.report({'INFO'}, "grabbing mode axis unlocked")
                context.area.tag_redraw()
            return {'RUNNING_MODAL'}
        if event.type in {'Y'}:
            if event.value == 'RELEASE':
                if self.axis_locked in (0, 1, 3):
                    self.axis_locked = 2
                    if context.window_manager.face2face.auto_view:
                        self.align_view2obj(context, modal)
                    modal.report({'INFO'}, "grabbing mode locked to y axis")
                else:
                    self.axis_locked = 0
                    modal.report({'INFO'}, "grabbing mode axis unlocked")
                context.area.tag_redraw()
            return {'RUNNING_MODAL'}
        if event.type in {'Z'}:
            if event.value == 'RELEASE':
                if self.axis_locked in (0, 1, 2):
                    self.axis_locked = 3
                    self.relative_pos = False
                    if context.window_manager.face2face.auto_view:
                        self.align_view2obj_z(context, modal)
                    modal.report({'INFO'}, "grabbing mode locked to z axis")
                else:
                    self.axis_locked = 0
                    modal.report({'INFO'}, "grabbing mode axis unlocked")
                    if context.window_manager.face2face.auto_view:
                        self.align_view2obj(context, modal)
                context.area.tag_redraw()
            return {'RUNNING_MODAL'}

    def handle_events(self, context, event, modal):
        """ handle events for the grabbing state """
        return_val = self.set_axis(context, event, modal)
        if return_val:
            self.grab_by_num_value(context, modal) 
            return return_val
        return_val=self.check_num_input(context, event, modal)
        if return_val: return return_val
        # if self.select_mode > 0:
        return_val=self.handle_events_pre(context, event, modal)
        if return_val: return return_val

        if event.type == 'MOUSEMOVE':

            if self.view_button_pressed:
                self.mouse_start=(event.mouse_region_x,
                                event.mouse_region_y)
                self.view_button_pressed=False

            self.grab_on_faces(context, event.mouse_region_x,
                               event.mouse_region_y, modal)
            # return {'RUNNING_MODAL'}

        if event.type in {'LEFTMOUSE', 'RET'}:

            if event.value == 'RELEASE' and self.legal_apply():
                modal.set_next_state("IDLE")
            return {'RUNNING_MODAL'}

        if event.type in {'RIGHTMOUSE', 'ESC'}:
            if event.value == 'RELEASE':
                self.reset(context, modal)
                modal.set_next_state("IDLE")
            return {'RUNNING_MODAL'}


        if event.type in {'R'}:
            if event.value == 'RELEASE':
                self.relative_pos = True
                self.limit_to_face = True
                # z axis only allowed in absolute positioning
                if self.axis_locked == 3:
                    self.axis_locked = 0
                self.reset(context, modal)
            return {'RUNNING_MODAL'}
        if event.type in {'A'}:
            if event.value == 'RELEASE':
                if self.relative_pos:
                    self.relative_pos = False
                    self.reset(context, modal)
                    return {'RUNNING_MODAL'}

        if event.type in {'L'}:
            if event.value == 'RELEASE':
                if not self.relative_pos:
                    self.limit_to_face = not self.limit_to_face
                    on_text = "ON" if self.limit_to_face else "OFF"
                    modal.report({'INFO'}, "Limit "+on_text)
                context.area.tag_redraw()
            return {'RUNNING_MODAL'}

        if event.type in {'A'}:
            if event.value == 'RELEASE':
                print("a press",self.relative_pos)
                if not self.relative_pos:
                    GrabState.adapt_offset(modal)
            return {'RUNNING_MODAL'}
    
        if event.type in {'O'}:
            if event.value == 'RELEASE':
                self.reset2start(modal)
            return {'RUNNING_MODAL'}

        return super().handle_events(context, event, modal)        


    def select_by_ngon(self, face_type, additive, modal, context, need_2b_selected=False):
        """ select objects of a specific target face type (tri,quad...) """ 
        
        if face_type == 1: # select all ...
            if not self.relative_pos:#... but only in absolute mode
                self.select_all(modal, context)
                return
            # else select only all of same ngon type of active
            face_type_active = mesh.get_tft_type(modal.projected_objects[modal.object_index], modal.target_object)
            face_type = face_type_active
        
        # no such type ... so do nothing
        if face_type not in modal.target_face_types:
            return
        
        if need_2b_selected: # select objects only if they were already selected before
            new_selected = [
                obj for obj in modal.projected_ngons[face_type] if obj in modal.selected_objects]

        # if not additive and in the special case of grabbing where in relative positioning mode all
        # former selected obj are going to be removed
        if not additive or self.relative_pos:#(isinstance(self, GrabState) and self.relative_pos):
             modal.selected_objects = set()
        if not need_2b_selected:
             new_selected = modal.projected_ngons[face_type]

        # update set and select the determed ones
        modal.selected_objects.update(new_selected)
        self.select_selected(modal, context)


    def set_target_face_type(self, context, modal, face_type_string):
        num = self.map_string2num(face_type_string)
        if num in modal.target_face_types:
            self.target_face_type = num
            self.reset(context, modal)
        if num == 1 and not self.relative_pos:
            self.target_face_type = num
            self.reset(context, modal)


    def legal_apply(self):
        """ checks if the current state of the move manipulation could be applied """
        # in relative mode everything must be inside ... it is if manipulator is
        # if self.relative_pos:
        #    return self.inside
        
        # if no limit is set everything is legal  
        if not self.limit_to_face:
            return True
        # in absolute mode everything must be inside ... it is if manipulator is an all others
        # inside = self.inside
        # for
        
        # Caution HACK !!!
        # check is done in draw.highlight_target_faces to avoid double calculation
        return self.inside

    def enter_state(self, context, event, modal):
        """ store the original matrix for canceling and the mouse pos at start """
        super().enter_state(context, event, modal)

        obj = modal.projected_objects[modal.object_index]
        self.target_face_type = mesh.get_num_of_verts(modal.target_object, obj.target_face_index)
        self.select_by_ngon(self.target_face_type, False, modal, context, True)

        self.mouse_start = (event.mouse_region_x,
                            event.mouse_region_y)

        if context.window_manager.face2face.auto_view:
            self.align_view2obj(context, modal)

        modal.set_status_text(" - GRABBING")


        self.limit_to_face = context.window_manager.face2face.start_with_limit
        self.relative_pos = context.window_manager.face2face.start_with_relative_pos
        if self.relative_pos: 
            self.limit_to_face = True

    def leave_state(self, modal):
        """ leaving the state cleanup  """


    def move_current_manipulator_obj(self, context, offset, modal):
        """ move the current active manipulator object by mapping
            the mouse movements from the view to the target face """
        obj = modal.projected_objects[modal.object_index]
        obj.select_set(True)

        offset_vec = mathutils.Vector((offset[0], offset[1], 0.0))

        # custom_transform = view.get_custom_transform_by_obj(context, obj)
        # org_local_rot = custom_transform.matrix
        org_local_rot = obj.matrix_custom_transform

        # get the view matrix
        # view_rot = context.region_data.view_matrix.decompose()[1].to_matrix()
        # view_rot.invert()
        r3d = context.area.spaces.active.region_3d
        view_rot = r3d.view_rotation.to_matrix()

        # calculate the difference between view roation and the target face one
        diff_view_rot = org_local_rot.to_quaternion().rotation_difference(
            view_rot.to_quaternion()).to_matrix()
        # diff_view_rot = local_rot.to_quaternion().rotation_difference(view_rot.to_quaternion()).to_matrix()

        # 2nd way could be to calculate difference between original and current rotation of
        # the object and use local matrix afterwards
        # local_rot = mathutils.Matrix(obj.matrix_local).decompose()[1].to_matrix()
        # local_rot = obj.matrix_local.to_quaternion()
        # diff_rot = org_local_rot.to_quaternion().rotation_difference(local_rot.to_quaternion()).to_matrix()

        # transform the mouse movement from view to target face orientation
        offset_vec = diff_view_rot @ offset_vec

        if self.axis_locked == 1:
            offset_vec.y = 0
            offset_vec.z = 0
        if self.axis_locked == 2:
            offset_vec.x = 0
            offset_vec.z = 0
        if self.axis_locked == 3:
            offset_vec.x = 0
            offset_vec.y = 0
        else:
            offset_vec.z = 0


        bpy.ops.transform.translate(
            value=(offset_vec.x, offset_vec.y, offset_vec.z),
            orient_matrix=org_local_rot)#,
            # constraint_axis=(True, False, False))

        """ bpy.ops.transform.translate(
            value=(offset[0], offset[1], 0.0),
            orient_type='LOCAL',
            orient_matrix_type='LOCAL',
            constraint_axis=(True, False, False)) """

        obj.select_set(False)  # ... and deselect afterwards

        return offset_vec



    def grab_on_faces(self, context, mouse_x, mouse_y, modal, given_offset=None):
        """ move all selected/projected objects along their individual target faces """
        if given_offset is None:
            offset = self.calc_offset(mouse_x, mouse_y)

        # deselect all objects ....
        for obj in modal.selected_objects:
            obj.select_set(False)

        # first move the actual active object
        if given_offset is None:
            offset_vec = self.move_current_manipulator_obj(context, offset, modal)
        else:
            offset_vec = given_offset

        # then move all other objects accordingly
        for obj in modal.selected_objects:
            # skip if its the actual active object but not if the offset is given
            if not modal.projected_objects[modal.object_index] is obj or given_offset is not None:
                # ...select only the current one to move them one by one
                obj.select_set(True)

                if self.relative_pos:
                    # set point releative to the actual active object relation to its target face
                    # TODO: make sure only object same ngon tf are seleceted
                    if mesh.faces_have_same_amount_of_verts(obj.target_face_index, 
                                                            modal.projected_objects[modal.object_index].target_face_index,
                                                            modal.target_object):
                        self.set_relative(
                            modal.projected_objects[modal.object_index]
                            # modal.selected_objects[modal.object_index]
                            , obj, modal)
                    else: obj.matrix_local = mathutils.Matrix(obj.matrix_local_org) # set to original pos
                else:
                    # set point to absolute position in relation to mouse movement
                    # custom_transform = view.get_custom_transform_by_obj(context, obj)
                    # org_local_rot = custom_transform.matrix
                    org_local_rot = obj.matrix_custom_transform
                    """ bpy.ops.transform.translate(
                        value=(offset_vec.x, offset_vec.y, offset_vec.z),
                        orient_matrix=org_local_rot) """
                    self.set_absolute(obj, org_local_rot, offset_vec)

                obj.select_set(False)  # ... and deselect afterwards

        # ...reselect the the formely deselected
        for obj in modal.selected_objects:
            obj.select_set(True)

    def set_absolute(self, obj, org_local_rot, offset_vec):
        offset_vec_relative = org_local_rot @ offset_vec
        obj.location = offset_vec_relative + obj.location


    def set_relative(self, current, obj, modal):
        """ move all selected/projected objects relative to the current manipulator object """
        target_face = modal.target_object.data.polygons[current.target_face_index]
        ctverts = mesh.get_verts_from_polygon(
            modal.target_object, target_face)

        ctverts_wc = mesh.converts_verts2wc(modal.target_object, ctverts)
        ctverts_wc = mesh.sort_high_z(ctverts_wc)

        # midpoint of the glueing active face of the current maipulator object
        current_mid_point = mesh.get_midpoint_by_active(current)

        tri_current = mesh.get_tri_from_point_inside_poly(
            ctverts_wc, current_mid_point)
        if not tri_current:
            return

        tri_current_i = mesh.get_indicies(tri_current, ctverts_wc)

        target_face_obj = modal.target_object.data.polygons[obj.target_face_index]
        otverts = mesh.get_verts_from_polygon(
            modal.target_object, target_face_obj)

        otverts_wc = mesh.converts_verts2wc(modal.target_object, otverts)
        otverts_wc = mesh.sort_high_z(otverts_wc)
        tri_object_wc = mesh.get_verts_by_indicies(tri_current_i, otverts_wc)

        projection_point_on_tri = mathutils.geometry.intersect_point_tri(
            current_mid_point, tri_current[0], tri_current[1], tri_current[2])

        if projection_point_on_tri is None:
            return

        # calculate projection point for object which need to be set to relative pos 
        obj_mid_point = mesh.get_midpoint_by_active(obj)
            
        tri_object_by_project = mesh.get_tri_from_point_inside_poly(
            otverts_wc, obj_mid_point)
        if not tri_object_by_project:
            return
            
        oprojection_point_on_tri = mathutils.geometry.intersect_point_tri(
            obj_mid_point, tri_object_by_project[0], tri_object_by_project[1], tri_object_by_project[2])

        if oprojection_point_on_tri is None:
            return

        # calculate the relative_point from object in respect to the projection point on target face from manipulator
        relative_point = mathutils.geometry.barycentric_transform(
            projection_point_on_tri,
            tri_current[0], tri_current[1], tri_current[2],
            tri_object_wc[0], tri_object_wc[1], tri_object_wc[2])
    
        obj.location = relative_point + (obj.location - oprojection_point_on_tri)

    def set_relative_ole(self, current, obj, modal):
        """ move all selected/projected objects relative to the current manipulator object """
        target_face = modal.target_object.data.polygons[current.target_face_index]
        ctverts = mesh.get_verts_from_polygon(
            modal.target_object, target_face)

        ctverts_wc = mesh.converts_verts2wc(modal.target_object, ctverts)
        ctverts_wc = mesh.sort_high_z(ctverts_wc)

        # midpoint of the glueing active face of the current maipulator object
        current_mid_point = mesh.get_midpoint_by_active(current)

        tri_current = mesh.get_tri_from_point_inside_poly(
            ctverts_wc, current_mid_point)
        if not tri_current:
            return

        tri_current_i = mesh.get_indicies(tri_current, ctverts_wc)

        target_face_obj = modal.target_object.data.polygons[obj.target_face_index]
        otverts = mesh.get_verts_from_polygon(
            modal.target_object, target_face_obj)

        otverts_wc = mesh.converts_verts2wc(modal.target_object, otverts)
        otverts_wc = mesh.sort_high_z(otverts_wc)
        tri_object_wc = mesh.get_verts_by_indicies(tri_current_i, otverts_wc)

        projection_point_on_tri = mathutils.geometry.intersect_point_tri(
            current_mid_point, tri_current[0], tri_current[1], tri_current[2])

        if projection_point_on_tri is None:
            return

        relative_point = mathutils.geometry.barycentric_transform(
            projection_point_on_tri,
            tri_current[0], tri_current[1], tri_current[2],
            tri_object_wc[0], tri_object_wc[1], tri_object_wc[2])

        obj.location = relative_point + \
            (obj.location - mesh.get_midpoint_by_active(obj))



    def calc_offset(self, mouse_x, mouse_y):
        """ calculate the offset to move objects by the mouse movement """
        diff_x = mouse_x - self.mouse_start[0]
        diff_y = mouse_y - self.mouse_start[1]
        offset = (diff_x/100.0, diff_y/100.0)
        self.mouse_start = (mouse_x, mouse_y)

        return offset


    def set_value_by_num(self, context, modal, num_string):
        if not super().set_value_by_num(context, modal, num_string):
            return self.leave_num_Enter_mode(modal, context, False)
        modal.set_status_text(" - GRABBING - KEY INPUT")
        # clamp angle to 360 degree max
        self.num_value = helpers.clamp(self.num_value, -100, 100)

        self.grab_by_num_value(context, modal)

    def grab_by_num_value(self, context, modal):
        
        offset_vec = mathutils.Vector((self.num_value,self.num_value,self.num_value))
        if self.axis_locked == 0:
            self.axis_locked = 1
        if self.axis_locked == 1:
            offset_vec.y = 0
            offset_vec.z = 0
        if self.axis_locked == 2:
            offset_vec.x = 0
            offset_vec.z = 0
        if self.axis_locked == 3:
            offset_vec.x = 0
            offset_vec.y = 0
        else:
            offset_vec.z = 0

        self.reset(context, modal)
        self.grab_on_faces(context, 0, 0, modal, offset_vec)

    def reset(self, context, modal):
        """ reset everything to its original state at the beginning of grab """
        # for obj in context.selected_objects:
        for obj in modal.projected_objects:
            obj.matrix_local = mathutils.Matrix(obj.matrix_local_org)

    def reset2start(self, modal):
        """ reset location of all projected objects to its original state at the beginning """
        for obj in modal.projected_objects:
            # recenter the face projected face to its target face
            mesh.position_mesh_face2face(obj, modal.target_object, obj.target_face_index)



