""" represents the Modal state where you can choose any other modal state or just select objects """
import bpy

if "bpy" in locals():
    import importlib
    from . import base
    importlib.reload(base)
    from . base import ModalState
    from . import blendout
    importlib.reload(blendout)
    from .. import mesh
    importlib.reload(mesh)
    from . import rotate
    importlib.reload(rotate)
    from . rotate import RotateState
    from . import scale
    importlib.reload(scale)
    from . scale import ScaleState
    from . import grab
    importlib.reload(grab)
    from . grab import GrabState
else:
    from . import base
    from . base import ModalState
    from . import blendout
    from .. import mesh
    from . import rotate
    from . rotate import RotateState
    from . import scale
    from . scale import ScaleState
    from . import grab
    from . grab import GrabState


class IdleState(ModalState):
    """ is the main modal state """
    name = "IDLE"

    @staticmethod
    def adapt_transformation(modal, context, event):
        """ adopt transformations from active to selected """
        rot_state = RotateState()
        rot_state.adapt_rotation(modal)
        context.view_layer.update()
        scale_state = ScaleState()
        scale_state.enter_state(context, event, modal)
        scale_state.adapt_scale(modal, context)
        GrabState.adapt_offset(modal)
        modal.set_status_text(" - IDLE")

    def handle_events(self, context, event, modal):
        """ handle events for the idle state """
        # if self.select_mode > 0:
        return_val = self.handle_events_pre(context, event, modal)
        if return_val:
            return return_val

        if event.type in {'LEFTMOUSE', 'RET'}:
            if event.value == 'RELEASE':
                # modal.remove_handlers(context)
                context.area.tag_redraw()
                # introduce an undo step
                # remove custom transforms
                mesh.delete_custom_transforms(context)
                bpy.ops.ed.undo_push(message="face2Face")
                # blendout.BlendOutState.text = "Successfully projected"
                modal.next_state = blendout.BlendOutState()
                modal.set_next_state(
                    "BLENDOUT",
                    blendout_text="Successfully applied",
                    blendout_ret_val='FINISHED')
                # modal.next_state.text_color = context.window_manager.face2face.status_text_error_color
                # modal.next_state.ret_val = 'FINISHED'
                # return {'FINISHED'}
            return {'RUNNING_MODAL'}

        if event.type in {'RIGHTMOUSE', 'ESC'}:
            if event.value == 'RELEASE':
                modal.next_state = blendout.BlendOutState()
                modal.set_next_state(
                    "BLENDOUT",
                    blendout_text="Aborted",
                    blendout_ret_val='CANCELLED',
                    blendout_color=context.window_manager.face2face.status_text_error_color)
                # modal.remove_handlers(context)
                context.area.tag_redraw()
                #draw.blendout = True
                # mesh.face2face_orientation_names = []
                # introduce an undo step ...
                bpy.ops.ed.undo_push()
                # ... so we can undo to the original state
                bpy.ops.ed.undo()

                # set activated back to true because props are now known to blender
                # context.window_manager.face2face.is_active = True

                # return {'CANCELLED'}
            return {'RUNNING_MODAL'}

        if event.type == 'S':
            if event.value == 'RELEASE':
                modal.set_next_state("SCALE")
                # modal.next_state = ScaleState()
                return {'RUNNING_MODAL'}
            return {'RUNNING_MODAL'}

        if event.type == 'G':
            if event.value == 'RELEASE':
                modal.set_next_state("MOVE")
                # modal.next_state = GrabState()
                return {'RUNNING_MODAL'}
            return {'RUNNING_MODAL'}

        if event.type == 'R':
            if event.value == 'RELEASE':
                modal.set_next_state("ROTATE")
                # modal.next_state = RotateState()
                return {'RUNNING_MODAL'}
            return {'RUNNING_MODAL'}

        if event.type in {'A'}:
            if event.value == 'RELEASE':
                IdleState.adapt_transformation(modal, context, event)
            return {'RUNNING_MODAL'}

        if event.type in {'O'}:
            if event.value == 'RELEASE':
                self.reset2start(modal)
                return {'RUNNING_MODAL'}
            return {'RUNNING_MODAL'}

        return super().handle_events(context, event, modal)

    def enter_state(self, context, event, modal):
        """ just set some parameter at the beginning of state """
        super().enter_state(context, event, modal)
        modal.set_status_text(" - IDLE")

    def leave_state(self, modal):
        """ leaving the state cleanup  """
