import mathutils
import math
import bpy


if "bpy" in locals():
    import importlib
    from .. import mesh
    importlib.reload(mesh)
    from .. import view
    importlib.reload(view)
    from . import base
    importlib.reload(base)
    from . base import ModalState
    from .. import helpers
    importlib.reload(helpers)
else:
    from .. import mesh
    from .. import view
    from . import base
    from . base import ModalState
    from .. import helpers


class RotateState(ModalState):
    """ is the modal state in which rotating and aligning takes place """
    name = "ROTATE"
    # mouse_start = (0, 0)

    setting_variation = False
    # the global variation at beginning of setting it
    variation_strength_start = 0
    # the global variation at entering the state
    variation_strength_start_state = 0

    def adapt_rotation(self, modal):
        """ adopt rotations from active to selected """
        active_obj = modal.projected_objects[modal.object_index]
        for obj in modal.selected_objects:
            obj.rotation = active_obj.rotation
            obj.rotation_vary = active_obj.rotation_vary
            self.rotate_on_faces(0, 0, modal, 0)

    def set_variation_strength(self, context, mouse_x, mouse_y, modal):
        """ scale all selected/projected objects on their individual target faces """
        diff_x = mouse_x - self.mouse_start[0]
        angle = diff_x/100.0
        self.mouse_start = (mouse_x, mouse_y)
        modal.rotation_variation_strength += angle
        modal.rotation_variation_strength = helpers.clamp(
            modal.rotation_variation_strength, 0, 1)
        context.area.tag_redraw()

    @staticmethod
    def set_per_object_variation(modal, context):
        """ shuffeling new random values per object """
        for obj in modal.projected_objects:
            obj.rotation_vary = helpers.get_random_rot_variation()
            if context.window_manager.face2face.vary_rotation_step > 0:
                rest = obj.rotation_vary % context.window_manager.face2face.vary_rotation_step
                obj.rotation_vary -= rest


    def set_per_object_variation_for(self, objects, context):
        """ shuffeling new random values per object """
        for obj in objects:
            obj.rotation_vary = helpers.get_random_rot_variation()
            if context.window_manager.face2face.vary_rotation_step > 0:
                rest = obj.rotation_vary % context.window_manager.face2face.vary_rotation_step
                obj.rotation_vary -= rest

    def zero_per_object_variation_for(self, objects):
        """ shuffeling new random values per object """
        for obj in objects:
            obj.rotation_vary = 0


    def handle_events_vary(self, context, event, modal):
        """ handle events for the scaling state in set variation mode """
        if event.type == 'MOUSEMOVE':

            if self.view_button_pressed:
                self.mouse_start=(event.mouse_region_x,
                                event.mouse_region_y)
                self.view_button_pressed=False

            self.set_variation_strength(context, event.mouse_region_x,
                                    event.mouse_region_y, modal)

            self.rotate_on_faces2angle(modal, True)
            return {'RUNNING_MODAL'}

        if event.type in {'LEFTMOUSE', 'RET'}:
            if event.value == 'RELEASE':
                self.setting_variation = False
                modal.report({'INFO'}, "Variation strength set to:" + \
                             str(modal.rotation_variation_strength))
                context.area.tag_redraw()
                modal.set_status_text(" - ROTATING")
                return {'RUNNING_MODAL'}

        if event.type in {'RIGHTMOUSE', 'ESC'}:
            if event.value == 'RELEASE':
                modal.rotation_variation_strength=self.variation_strength_start
                self.setting_variation=False
                self.rotate_on_faces2angle(modal, True)
                modal.report({'INFO'}, "setting variation strength cancelled")
                modal.set_status_text(" - ROTATING")
                context.area.tag_redraw()
                return {'RUNNING_MODAL'}

        if event.type in {'V'}:
            if event.value == 'RELEASE':
                self.set_per_object_variation_for(modal.selected_objects, context)
                self.rotate_on_faces2angle(modal, True)
                modal.report(
                    {'INFO'}, "New random variation for selected objects shuffled")
                return {'RUNNING_MODAL'}

        if event.type in {'Z'}:
            if event.value == 'RELEASE':
                self.zero_per_object_variation_for(modal.selected_objects)
                self.rotate_on_faces2angle(modal, True)
                modal.report(
                    {'INFO'}, "random variation set to zero for selected objects")
                return {'RUNNING_MODAL'}

    def handle_events_normal(self, context, event, modal):
        """ handle events for the scaling state in normal mode """
        # if self.select_mode > 0:
        return_val = self.check_num_input(context, event, modal)
        if return_val: return return_val

        return_val = self.handle_events_pre(context, event, modal)
        if return_val: return return_val

        if event.type == 'MOUSEMOVE':

            if self.view_button_pressed:
                self.mouse_start= (event.mouse_region_x,
                                event.mouse_region_y)
                self.view_button_pressed= False

            self.rotate_on_faces(event.mouse_region_x,
                                 event.mouse_region_y, modal)
            # return {'RUNNING_MODAL'}

        if event.type in {'LEFTMOUSE', 'RET'}:

            if event.value == 'RELEASE':
                modal.set_next_state("IDLE")
            return {'RUNNING_MODAL'}

        if event.type in {'RIGHTMOUSE', 'ESC'}:
            if event.value == 'RELEASE':
                self.reset(context, modal)
                modal.set_next_state("IDLE")
            return {'RUNNING_MODAL'}

        if event.type in {'V'}:
            if event.value == 'RELEASE':
                if not self.setting_variation:
                    self.variation_strength_start= modal.rotation_variation_strength
                    self.setting_variation= True
                    modal.report({'INFO'}, "Var set :"+ \
                                 str(modal.rotation_variation_strength))
                    modal.set_status_text(" - SET RANDOM ROTATION")
                    context.area.tag_redraw()
            return {'RUNNING_MODAL'}

        if event.type in {'A'}:
            if event.value == 'RELEASE':
                self.adapt_rotation(modal)
            return {'RUNNING_MODAL'}

        if event.type in {'O'}:
            if event.value == 'RELEASE':
                self.reset2start(modal)
            return {'RUNNING_MODAL'}

        if event.type in {'I'}:
            if event.value == 'RELEASE':
                 self.toggle_flipped(modal.selected_objects, modal, context)
            return {'RUNNING_MODAL'}

        if event.type in {'E'}:
            if event.value == 'RELEASE':
                if not event.shift:
                    self.rotate_on_faces2align_obj_edge2target_z(modal, True)
                else:
                    self.rotate_on_faces2align_obj_edge2target_z(modal, False)
                if event.alt:
                    if not event.shift:
                        self.rotate_on_faces2align_obj_first_edge2target_current(
                            modal, True)
                    else:
                        self.rotate_on_faces2align_obj_first_edge2target_current(
                            modal, False)
            return {'RUNNING_MODAL'}

        if event.type in {'RIGHT_ARROW'}:
            if event.value == 'RELEASE':
                if event.shift:
                    self.rotate_on_faces(0, 0, modal, math.radians(context.window_manager.face2face.shift_rotation_step))
                if event.ctrl:
                    self.rotate_on_faces(0, 0, modal, math.radians(context.window_manager.face2face.ctrl_rotation_step))
                if event.alt:
                    self.rotate_on_faces(0, 0, modal, math.radians(context.window_manager.face2face.alt_rotation_step))
                if not event.shift and not event.alt and not event.ctrl:
                    self.rotate_on_faces(0, 0, modal, math.radians(context.window_manager.face2face.normal_rotation_step))
            return {'RUNNING_MODAL'}
        if event.type in {'LEFT_ARROW'}:
            if event.value == 'RELEASE':
                if event.shift:
                    self.rotate_on_faces(0, 0, modal, math.radians(-context.window_manager.face2face.shift_rotation_step))
                if event.ctrl:
                    self.rotate_on_faces(0, 0, modal, math.radians(-context.window_manager.face2face.ctrl_rotation_step))
                if event.alt:
                    # self.reset(context, modal)
                    self.rotate_on_faces(0, 0, modal, math.radians(-context.window_manager.face2face.alt_rotation_step))
                if not event.shift and not event.alt and not event.ctrl:
                    self.rotate_on_faces(0, 0, modal, math.radians(-context.window_manager.face2face.normal_rotation_step))
            return {'RUNNING_MODAL'}


    def handle_events(self, context, event, modal):
        """ handle events for the rotate state """

        if self.setting_variation:
            return_val = self.handle_events_vary(context, event, modal)
            if return_val: return return_val

        if not self.setting_variation:
            return_val = self.handle_events_normal(context, event, modal)
            if return_val: return return_val

        return super().handle_events(context, event, modal)


    def enter_state(self, context, event, modal):
        """ store the original matrix for canceling and the mouse pos at start """
        super().enter_state(context, event, modal)
        for obj in modal.projected_objects:
            obj.scaling_center = mesh.get_midpoint_by_active(obj)
            obj.rotation_org = obj.rotation
            obj.rotation_vary_org = obj.rotation_vary

        self.variation_strength_start_state = modal.rotation_variation_strength
        self.mouse_start = (event.mouse_region_x,
                             event.mouse_region_y)
        modal.set_status_text(" - ROTATING")


    def rotate_on_faces_to_next_edge_bothz(self, modal):
        """ rotate all selected/projected objects around their local z axis """
        # deselect all objects ....
        # for obj in modal.projected_objects:
        for obj in modal.selected_objects:
            obj.select_set(False)

        # for obj in modal.projected_objects:
        for obj in modal.selected_objects:
            # skip if wrong amount of verts when tri, quad or specific ngon only is ON
            # if not self.obj_has_current_tft(obj, modal):
            #    continue

            target_face = modal.target_object.data.polygons[obj.target_face_index]

            ctverts = mesh.get_verts_from_polygon(modal.target_object, target_face)
            ctverts_wc = mesh.converts_verts2wc(modal.target_object, ctverts)
            ctverts_wc = mesh.sort_high_z(ctverts_wc)
            z_edge_target = mesh.get_highest_z_edge(ctverts_wc)

            sverts = mesh.get_verts_from_polygon(obj, mesh.get_active_polygon(obj) )
            sverts_wc = mesh.converts_verts2wc(obj, sverts)
            sverts_wc = mesh.sort_high_z(sverts_wc)
            z_edge_source = mesh.get_highest_z_edge(sverts_wc)

            z_edge_source_vec = z_edge_source[1]-z_edge_source[0]
            z_edge_target_vec = z_edge_target[0]-z_edge_target[1]

            self.rotate_on_faces_align_edges(obj, z_edge_source_vec, z_edge_target_vec, modal)

        # delete variation ... because makees nos sense for aligned ones(TODO check if it could be left for not selectec)
        # modal.rotation_variation_strength = 0

        # ...reselect the the formely deselected
        # for obj in modal.projected_objects:
        for obj in modal.selected_objects:
            obj.select_set(True)


    def get_next_target_edge(self, obj, modal):
        """ get next obj edge to objects currrently aligned edge """
        target_face = modal.target_object.data.polygons[obj.target_face_index]
        tverts = mesh.get_verts_from_polygon(modal.target_object, target_face)       
        obj.tf_aligned_edge_index += 1
        if obj.tf_aligned_edge_index >= len(tverts):
            obj.tf_aligned_edge_index = 0
        return self.get_obj_edge(modal.target_object, tverts, obj.tf_aligned_edge_index)

    def get_prev_target_edge(self, obj, modal):
        """ get prev obj edge to objects currrently aligned edge """
        target_face = modal.target_object.data.polygons[obj.target_face_index]
        tverts = mesh.get_verts_from_polygon(modal.target_object, target_face)
        obj.tf_aligned_edge_index -= 1
        if obj.tf_aligned_edge_index < 0:
            obj.tf_aligned_edge_index = len(tverts)-1
        return self.get_obj_edge(modal.target_object, tverts, obj.tf_aligned_edge_index)

    def get_next_obj_edge(self, obj):
        """ get next obj edge to objects currrently aligned edge """
        sverts = mesh.get_verts_from_polygon(obj, mesh.get_active_polygon(obj) )
        obj.aligned_edge_index += 1
        if obj.aligned_edge_index >= len(sverts):
            obj.aligned_edge_index = 0
        return self.get_obj_edge(obj, sverts, obj.aligned_edge_index)

    def get_prev_obj_edge(self, obj):
        """ get prev obj edge to objects currrently aligned edge """
        sverts = mesh.get_verts_from_polygon(obj, mesh.get_active_polygon(obj) )
        obj.aligned_edge_index -= 1
        if obj.aligned_edge_index < 0:
            obj.aligned_edge_index = len(sverts)-1
        return self.get_obj_edge(obj, sverts, obj.aligned_edge_index)

    def get_obj_edge(self, obj, verts, index):
        """ get obj edge by vert index counterclockwise """
        verts_wc = mesh.converts_verts2wc(obj, verts)
        edge = mesh.get_counterclockwise_edge(verts_wc[index], verts_wc)
        return edge


    def rotate_on_faces2align_obj_first_edge2target_current(self, modal, clockwise):
        """ rotate so that the first objects edge aligns with
            its target face currrently aligned edge """

        # deselect all objects ....
        # for obj in modal.projected_objects:
        for obj in modal.selected_objects:
            obj.select_set(False)

        # for obj in modal.projected_objects:
        for obj in modal.selected_objects:
            # skip if wrong amount of verts when tri, quad or specific ngon only is ON
            # if not self.obj_has_current_tft(obj, modal):
            #    continue

            if clockwise:
                z_edge_target = self.get_prev_target_edge(obj, modal)
            else:
                z_edge_target = self.get_next_target_edge(obj, modal)

            sverts = mesh.get_verts_from_polygon(obj, mesh.get_active_polygon(obj) )
            z_edge_source = self.get_obj_edge(obj, sverts, 0)

            z_edge_source_vec = z_edge_source[1]-z_edge_source[0]
            z_edge_target_vec = z_edge_target[0]-z_edge_target[1]

            self.rotate_on_faces_align_edges(obj, z_edge_source_vec, z_edge_target_vec, modal)

        # delete variation ... because makees nos sense for aligned ones(TODO check if it could be left for not selectec)
        # modal.rotation_variation_strength = 0

        # ...reselect the the formely deselected
        # for obj in modal.projected_objects:
        for obj in modal.selected_objects:
            obj.select_set(True)

    def rotate_on_faces2align_obj_edge2target_z(self, modal, clockwise):
        """ rotate so that next/prev edge to objects currrently aligned edge aligns with 
            its target face highest z edge """
            
        # deselect all objects ....
        # for obj in modal.projected_objects:
        for obj in modal.selected_objects:
            obj.select_set(False)

        # for obj in modal.projected_objects:
        for obj in modal.selected_objects:
            # skip if wrong amount of verts when tri, quad or specific ngon only is ON
            # if not self.obj_has_current_tft(obj, modal):
            #    continue

            target_face = modal.target_object.data.polygons[obj.target_face_index]
            ctverts = mesh.get_verts_from_polygon(modal.target_object, target_face)
            ctverts_wc = mesh.converts_verts2wc(modal.target_object, ctverts)
            ctverts_wc = mesh.sort_high_z(ctverts_wc)
            z_edge_target = mesh.get_highest_z_edge(ctverts_wc)
              
            if clockwise:
                z_edge_source = self.get_prev_obj_edge(obj)
            else:
                z_edge_source = self.get_next_obj_edge(obj)

            z_edge_source_vec = z_edge_source[1]-z_edge_source[0]
            z_edge_target_vec = z_edge_target[0]-z_edge_target[1]

            self.rotate_on_faces_align_edges(obj, z_edge_source_vec, z_edge_target_vec, modal)
        
        # delete variation ... because makees nos sense for aligned ones(TODO check if it could be left for not selectec)
        # modal.rotation_variation_strength = 0

        # ...reselect the the formely deselected
        # for obj in modal.projected_objects:
        for obj in modal.selected_objects:
            obj.select_set(True)

    def rotate_on_faces_align_edges(self, obj, source_edge, target_edge, modal):
        """ rotate all selected/projected objects so that they are aligned by edges """
        mat_rot = obj.matrix_local_startorg.decompose()[1].to_matrix().to_4x4()
        mat_rot.invert()

        source_edge = mat_rot @ source_edge
        target_edge = mat_rot @ target_edge
        source_edge_2d = mathutils.Vector((source_edge.x, source_edge.y))
        target_edge_2d = mathutils.Vector((target_edge.x, target_edge.y))
        angle = -source_edge_2d.angle_signed(target_edge_2d)
        
        self.rotate_on_face_by_angle(angle, obj)
        # add variation to the angle ... because we will delete strength after all
        # print("obj.rotation bv",math.degrees(obj.rotation), "  angle",math.degrees(angle), "vary",math.degrees(obj.rotation_vary * modal.rotation_variation_strength))
        obj.rotation += angle + (obj.rotation_vary * modal.rotation_variation_strength)
        # print("obj.rotation af",math.degrees(obj.rotation))
        obj.rotation_vary = 0
        
    def rotate_on_face_by_angle(self, angle, obj):
        """ rotate a single selected/projected object around its local z axis 
             by an given angle in radians """
        # ...select only the current one to rotate them one by one
        obj.select_set(True)
        bpy.context.view_layer.objects.active = obj
 
        bpy.ops.transform.rotate(
            value=angle,
            orient_type='LOCAL',
            orient_matrix_type='LOCAL',
            center_override=(obj.scaling_center),
            constraint_axis=(False, False, True))
 
        obj.select_set(False)  # ... and deselect afterwards


    angle_change = 0
    ANGLE_TRESHOLD = 0.09
    def rotate_on_faces(self, mouse_x, mouse_y, modal, angle=None):
        """ rotate all selected/projected objects around their local z axis """
        # calculate angle by mouse movement
        if angle is None:
            angle = self.calc_angle(mouse_x, mouse_y)

        self.add_angle2objs(angle, modal)#, all_objects)
        self.angle_change = self.angle_change + angle

        # if self.angle_change >= self.ANGLE_TRESHOLD or \
        #   self.angle_change <= -self.ANGLE_TRESHOLD :
        #    self.angle_change = 0
        #    self.rotate_on_faces2angle(modal)
        self.rotate_on_faces2angle(modal)

    def rotate_on_faces2angle(self, modal, all_objects=False):
        """ rotate all selected/projected objects around their local z axis
            by an given angle in radians """
        objects = modal.projected_objects if all_objects else modal.selected_objects

        for obj in objects:
                       
            angle = obj.rotation + obj.rotation_vary * modal.rotation_variation_strength
            self.rotate_on_face2angle(angle, obj)
    
        bpy.context.view_layer.objects.active = modal.projected_objects[modal.object_index]
        
    def toggle_flipped(self, objects, modal, context):
        """ toggle rotation direction """
        for obj in objects:
            obj.rotation_flipped = False if obj.rotation_flipped else True

        #draw.update_rotflip_objects(modal.selected_objects)
        context.window_manager.f2f_tmp_vars.update_rotflip_objects = True
        context.window_manager.f2f_tmp_vars.rotflip_objects = []
        context.window_manager.f2f_tmp_vars.rotflip_objects += modal.selected_objects


    def add_angle2objs(self, angle_start, modal):
        """ add angle to stored angle in selected obj """
        for obj in modal.selected_objects:
            if obj.rotation_flipped:
                angle = angle_start * -1
            else:
                angle = angle_start
                
            obj.rotation += angle

    def set_obj_angle(self, obj, angle):
        """ set stored angle in obj to specific value """
        obj.rotation = angle


    def rotate_on_face2angle(self, angle, obj):
        """ rotate all given objects around their local z axis
            by an given angle in radians """
        # get rotation matrix from original start transform matrix + rotation on local z
        mat_rot = obj.matrix_local_startorg.decompose()[1].to_matrix().to_4x4()   
        mat_rot = mat_rot @ mathutils.Matrix.Rotation(-angle, 4, 'Z')     

        # get location and scale out of the current transform matrix
        deco = obj.matrix_local.decompose()  
        mat_loc = mathutils.Matrix.Translation(deco[0])
        mat_sca = mesh.get_sca_matrix(deco[2])

        # store mid of projection face, object pos and local matrix bevor altering
        org_mid = mesh.get_midpoint_by_active(obj)
        obj_loc_org = obj.location.copy()
        matrix_local_bv = obj.matrix_local.copy()

        # calculate offset to mid in world coords
        local_mid = org_mid - obj.location    

        # create new transform matrix from original rotation and current loc,scale
        obj.matrix_local = mat_loc.copy() @ mat_rot @ mat_sca

        # now projected face mid has changed ... fix that -->
        # transform offset to local mid into its original state
        matrix_local_bv_rot = matrix_local_bv.decompose()[1].to_matrix()
        matrix_local_bv_rot.invert()
        local_mid = matrix_local_bv_rot @ local_mid

        # transform original local mid into altered current transform
        new_mid = mat_rot @ local_mid

        # calculate the offset of the object through the rotation and put it back
        new_mid = obj_loc_org + new_mid.copy() 
        loc_diff = org_mid - new_mid
        obj.location += loc_diff


    def calc_angle(self, mouse_x, mouse_y):
        """ calculate the offset to move objects by the mouse movement """
        diff_x = mouse_x - self.mouse_start[0]
        # diff_y = mouse_y - self.mouse_start[1]
        angle = diff_x/100.0
        self.mouse_start = (mouse_x, mouse_y)
        
        return angle


    def set_value_by_num(self, context, modal, num_string):
        if not super().set_value_by_num(context, modal, num_string):
            return self.leave_num_Enter_mode(modal, context, False)
        modal.set_status_text(" - GRABBING - KEY INPUT")
        # clamp angle to 360 degree max
        self.num_value = helpers.clamp(self.num_value, -360, 360)
        
        self.reset(context, modal)
        self.rotate_on_faces(0, 0, modal, math.radians(self.num_value))


    def reset(self, context, modal):
        """ reset everything to its original state at the beginning of rotate """
        # for obj in context.selected_objects:
        for obj in modal.projected_objects:
            obj.matrix_local = mathutils.Matrix(obj.matrix_local_org)
            # update the stored rotation per object
            obj.rotation = obj.rotation_org
            obj.rotation_vary = obj.rotation_vary_org
            # set back the global rotation variation strength
        modal.rotation_variation_strength = self.variation_strength_start_state


    def reset2start_by_objects(self, modal, objects, reset_obj_rotation=True):
        """ reset rotation of all projected objects to its original state at the beginning """
        for obj in objects:

            # custom_transform = view.get_custom_transform_by_obj(bpy.context, obj)
            # org_local_rot = custom_transform.matrix.copy()
            org_local_rot = obj.matrix_custom_transform.copy()
            org_local_rot.invert()

            # rb0 = obj.matrix_local.decompose()[1].to_matrix().to_4x4()  @ org_local_rot.to_4x4()    

            # get rotation matrix from original start transform matrix
            mat_rot = obj.matrix_local_startorg.decompose()[1].to_matrix().to_4x4()    
            
            # rb = obj.matrix_local_startorg.decompose()[1].to_matrix().to_4x4()  @ org_local_rot.to_4x4()       

            # get location and scale out of the current transform matrix
            deco = obj.matrix_local.decompose()  
            mat_loc = mathutils.Matrix.Translation(deco[0])
            # vec_sca = deco[2]
            mat_sca = mesh.get_sca_matrix(deco[2])

            # store mid of projection face bevor resetting
            org_mid = mesh.get_midpoint_by_active(obj)

            # create new transform matrix from original rotation and current loc,scale
            obj.matrix_local = mat_loc.copy() @ mat_rot @ mat_sca

            # the rotation sets the object off because it is not rotated around origin
            # ...so we need to set it back into its place
            bpy.context.view_layer.update()
            new_mid = mesh.get_midpoint_by_active(obj)
            loc_diff = org_mid - new_mid
            obj.location += loc_diff
            bpy.context.view_layer.update()


            # the rotation sets the object off because it is not rotated around origin
            # ...so we need to set it back into its place
            bpy.context.view_layer.update()

            # reset the stored angle
            if reset_obj_rotation:
                obj.rotation = 0 
                modal.rotation_variation_strength = 0
            


    def reset2start(self, modal):
        """ reset rotation of all projected objects to its original state at the beginning """
        self.reset2start_by_objects(modal, modal.projected_objects)

