""" File with the addon ui panels """
import bpy


if "bpy" in locals():
    import importlib
    from .. import config
    importlib.reload(config)
    from ..config import settings
    importlib.reload(settings)
    from .. import helpers
    importlib.reload(helpers)
    from ..helpers import addon_prefs
    from . import hotkey_display
    importlib.reload(hotkey_display)
    from .. import states
    importlib.reload(states)
    #from . import projection_panel
    # importlib.reload(projection_panel)
else:
    #from . import projection_panel
    from ..config import settings
    from .. import helpers
    from ..helpers import addon_prefs
    from . import hotkey_display
    from .. import states


class VIEW3D_PT_f2f_options_panel(bpy.types.Panel):

    bl_idname = 'VIEW3D_PT_f2f_options_panel'
    bl_category = 'f2F'
    bl_label = 'face2FACE SETTINGS'
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'

    def draw(self, context):
        layout = self.layout

        if context.window_manager.face2face.is_modal_active:
            self.draw_active(context)
        else:
            self.draw_inactive(context)

    def draw_inactive(self, context):

        layout = self.layout
        col = layout.column()
        col.label(text="Local settings per file:")
        # draw local settings group
        # context.window_manager.face2face.draw(
        #    context, layout, "Local settings per file:")

    def draw_blendout(self, context):
        layout = self.layout
        col = layout.column()

        if not hotkey_display.state.ret_val == 'FINISHED':
            col.alert = True

        col.label(text=""+hotkey_display.state.text)

    def draw_msg(self, msg):
        layout = self.layout
        col = layout.column()

        col.label(text=""+msg)

    def draw_active(self, context):

        if context.window_manager.f2f_tmp_vars.options_panel_msg:
            return self.draw_msg(
                context.window_manager.f2f_tmp_vars.options_panel_msg)

        if isinstance(hotkey_display.state, states.BlendOutState):
            return self.draw_blendout(context)

        layout = self.layout
        # show active projection
        box_o = layout.box()
        #box_o.label(text="Modal is activly working on:")
        box = box_o.box()
        box.label(text=" - "+settings.current_projection_name+" - ")

        # context help
        if hotkey_display.state:
            hotkey_display.draw2layout(layout, context)
            hotkey_display.set_selction_keys_layout(layout, context)


def register():
    bpy.utils.register_class(VIEW3D_PT_f2f_options_panel)


def unregister():
    bpy.utils.unregister_class(VIEW3D_PT_f2f_options_panel)
